QUOTE = '"'
QUASIQUOTE = '`'
UNQUOTE = ','

std_env = {
    '+': sum,
    'list': list,
    'let': {'macro': 'let'},
    'fun': {'macro': 'lambda'},
    'macro': {'macro': 'macro'},
    'select': {'macro': 'select'},
    'eval': {'macro': 'eval'},
    'case': {'macro': 'case'},
    'true': [[]],  # just empty binding
    'false': [],  # no bindings
    QUOTE: {'macro': 'quote'},
    QUASIQUOTE: {'macro': 'quasiquote'},
}


def run(exprs, env=std_env):
    env = dict(env)
    env_stack = []
    value_stack = []
    command_stack = list(exprs)
    while command_stack:
        match command_stack.pop():
            case int(num):
                value_stack.append(num)
            case str(ident):
                value_stack.append(env[ident])
            case list([]):
                value_stack.append([])
            case list([func, *args]):
                command_stack.append({'apply': args})
                command_stack.append(func)
            case {'apply': args}:
                # when 'apply' runs, func eval'd onto val stack
                # args not eval'd
                match value_stack.pop(), args:
                    case func, args if callable(func):
                        command_stack.append(
                            {'call': func, 'arity': len(args)})
                        for arg in args:
                            command_stack.append(arg)
                    case {
                        'params': patterns,
                        'lambda_body': _,
                        'env': dict(_),
                    } as func, args:
                        # lambda
                        if len(args) != len(patterns):
                            raise ValueError(
                                f'expected parameters {patterns}, found args {args}'
                            )
                        command_stack.append(
                            {'call': func, 'arity': len(args)})
                        for arg in args:
                            command_stack.append(arg)
                    case {
                        'params': patterns,
                        'macro_body': _,
                        'env': dict(_),
                    } as func, args:
                        # user macro
                        if len(args) != len(patterns):
                            raise ValueError(
                                f'expected parameters {patterns}, found args {args}'
                            )
                        # eval result of macro
                        command_stack.append(
                            {'call': {'macro': 'eval'}, 'arity': 1})
                        command_stack.append(
                            {'call': func, 'arity': len(args)})
                        while args:
                            # reverse order, args never popped from cmd stack
                            value_stack.append(args.pop())
                    case {'macro': 'eval'} as func, args:
                        command_stack.append(
                            {'call': func, 'arity': len(args)})
                        for arg in args:
                            command_stack.append(arg)
                    case {'macro': 'quote'}, [arg]:
                        value_stack.append(arg)
                    case {'macro': 'quasiquote'}, [arg]:
                        match arg:
                            case list([quasiquote, quoted] as arg) if quasiquote == QUASIQUOTE:
                                value_stack.append(arg)
                            case list([unquote, unquoted]) if unquote == UNQUOTE:
                                command_stack.append(unquoted)
                            case list(quoted_list):
                                command_stack.append(
                                    [
                                        'list',
                                        *(
                                            [QUASIQUOTE, element]
                                            for element in quoted_list
                                        ),
                                    ]
                                )
                            case _:
                                value_stack.append(arg)
                    case {'macro': 'let'}, [*patterns, body]:
                        command_stack.append({'restore_env': None})
                        command_stack.append(body)
                        for key, value in patterns:
                            command_stack.append({'set': key})
                            command_stack.append(value)
                        command_stack.append({'enter_env': {}})
                    case {'macro': 'lambda'}, [patterns, body]:
                        value_stack.append(
                            {'params': patterns, 'lambda_body': body,
                                'env': dict(env)}
                        )
                    case {'macro': 'macro'}, [patterns, body]:
                        value_stack.append(
                            {'params': patterns, 'macro_body': body,
                                'env': dict(env)}
                        )
                    case {'macro': 'select'}, list(
                        [(condition, body), *clauses, otherwise]
                    ):
                        command_stack.append(
                            {'select': clauses, 'body': body,
                                'otherwise': otherwise}
                        )
                        command_stack.append(condition)
                    case {'macro': 'case'}, list([*patterns, expr] as args):
                        command_stack.append(
                            {'call': {'macro': 'case'}, 'arity': len(args)})
                        for pattern in patterns:
                            value_stack.append(pattern)
                        command_stack.append(expr)
                    case {'macro': name}, args:
                        raise TypeError(
                            f'expected {name} macro parameters, found {args}'
                        )
                    case _:
                        raise TypeError(f'expected callable, found {func}')
            case {'call': func, 'arity': int(arity)}:
                # func eval'd, args eval'd
                args = [value_stack.pop() for _ in range(arity)]
                match func:
                    case {
                        'params': patterns,
                        'lambda_body': body,
                        'env': dict(saved_env),
                    }:
                        # lambda
                        command_stack.append({'restore_env': None})
                        command_stack.append(
                            {
                                'apply': [
                                    *([pattern, [QUOTE, arg]]
                                      for pattern, arg in zip(patterns, args)),
                                    body,
                                ]
                            }
                        )
                        value_stack.append({'macro': 'let'})
                        command_stack.append({'enter_env': saved_env})
                    case {
                        'params': patterns,
                        'macro_body': body,
                        'env': dict(saved_env),
                    }:
                        # user macro
                        command_stack.append({'restore_env': None})
                        command_stack.append(
                            {
                                'apply': [
                                    *([pattern, [QUOTE, arg]]
                                      for pattern, arg in zip(patterns, args)),
                                    body,
                                ]
                            }
                        )
                        value_stack.append({'macro': 'let'})
                        command_stack.append({'enter_env': saved_env})
                    case {'macro': 'eval'}:
                        for arg in args:
                            command_stack.append(arg)
                    case {'macro': 'case'}:
                        [expr, *patterns] = args
                        command_stack.append(
                            {'cases': 'or', 'arity': len(patterns)})
                        patterns.reverse()
                        for pattern in patterns:
                            command_stack.append(
                                {'case': pattern, 'expr': expr})
                    case func if callable(func):
                        value_stack.append(func(args))
                    case func:
                        raise ValueError(f'expected call, found {func}')
            case {'select': clauses, 'body': body, 'otherwise': otherwise}:
                match value_stack.pop():
                    case list([bindings, *_]):
                        # success
                        command_stack.append({'restore_env': None})
                        command_stack.append(body)
                        command_stack.append(
                            {'enter_env': {key: value for [
                                key, value] in bindings}}
                        )
                    case _:
                        match clauses:
                            case []:
                                # no clauses left
                                command_stack.append(otherwise)
                            case [(condition, body), *clauses]:
                                command_stack.append(
                                    {
                                        'select': clauses,
                                        'body': body,
                                        'otherwise': otherwise,
                                    }
                                )
                                command_stack.append(condition)
            case {'case': pattern, 'expr': expr}:
                match pattern, expr:
                    case list([unquote, str(ident)]), expr if unquote == UNQUOTE:
                        # can bind wildcard
                        value_stack.append([[[ident, expr]]])
                    case int(pattern), int(expr) if pattern == expr:
                        value_stack.append([[]])
                    case list(patterns), list(exprs) if len(patterns) == len(exprs):
                        # bind elements
                        command_stack.append(
                            {'cases': 'and', 'arity': len(patterns)})
                        for pattern, expr in zip(patterns, exprs):
                            command_stack.append(
                                {'case': pattern, 'expr': expr})
                    case _:
                        value_stack.append([])
            case {'cases': 'and', 'arity': arity}:
                patterns = [value_stack.pop() for _ in range(arity)]
                matches = [[]]
                for pattern in patterns:
                    new_matches = []
                    for match in matches:
                        for option in pattern:
                            new_match = match + option
                            idents = [ident for ident, _ in new_match]
                            if len(idents) == len(set(idents)):
                                new_matches.append(new_match)
                    matches = new_matches
                value_stack.append(matches)
            case {'cases': 'or', 'arity': arity}:
                patterns = [value_stack.pop() for _ in range(arity)]
                value_stack.append(
                    [pattern for match in patterns for pattern in match])
            case {'enter_env': new_env}:
                env_stack.append(dict(env))
                env |= new_env
            case {'restore_env': None}:
                env = env_stack.pop()
            case {'set': key}:
                env[key] = value_stack.pop()
            case command:
                raise ValueError(f'expected command, found {command}')
    return value_stack


def read(s):
    o = [[]]
    for symbol in '()[]`,"':
        s = s.replace(symbol, f' {symbol} ')
    for t in s.split():
        match t:
            case '(' | '[':
                o.append([])
            case ')' | ']':
                tmp = o.pop()
                o[-1].append(tmp)
            case str(num) if num.isnumeric():
                o[-1].append(int(num))
            case str(ident):
                o[-1].append(ident)
    [o] = o
    return o


tutorials = [
    ('(+ 1 2 3)', 6),
    ('(let (a 1) (b 2) (c 3) (d 4) b)', 2),
    ('((fun (x) x) 0)', 0),
    ('((fun (a b c d) b) 1 2 3 4)', 2),
    ('(select (false 1) (true 2) (true 3) 4)', 2),
    ('(select (false 1) (false 2) (false 3) 4)', 4),
    ('true', [[]]),
    ('false', []),
    ('(case 1 (+ 1 1))', []),
    ('(case 2 (+ 1 1))', [[]]),
    ('(case (, x) (+ 1 1))', [[['x', 2]]]),
    ('(case [[(,x) (,y) 3] (,z)] (" ((1 2 3) (4 5 6))))',
     [[['x', 1], ['y', 2], ['z', [4, 5, 6]]]]),
    ('(select ((case (, x) 1) (+ x 2)) 0)', 3),
    ('(select ((case ((, x) (, y)) (" (1 2))) (+ x y)) 0)', 3),
    ('(let (no []) (select (no 0) 1))', 1),
    ('(let (bind-a (` [[(a (, 0))]])) (select (bind-a a) 1))', 0),
    ('(case 1 2 3 (,x) (,y) (+ 1 1))', [[], [['x', 2]], [['y', 2]]]),
    ('(` (1 2 (+ 3 4) (, (+ 5 6))))', [1, 2, ['+', 3, 4], 11]),
    ('(` (` (, (+ 1 1))))', ['`', [',', ['+', 1, 1]]]),
    ('((fun () (" (+ 1 1))))', ['+', 1, 1]),
    ('((macro () (" (+ 1 1))))', 2),
    ('((macro (x) (` (+ 2 (, x)))) 1)', 3),
    ('((macro (x y z) (` ((,x) (,y) (,z)))) + 2 3)', 5),
    ('(let (if (macro (c t f) (` (select ((, c) (, t)) (, f))))) (if true 1 2))', 1),
    ('(let (x 0) (select [(case 0 x) (" zero)] [(case 1 x)] (" one) (" big)))', 'zero'),
    ('(let (car (fun (pair) (select [(case [(,car) (,cdr)] pair) car] (" error)))) (car (" [1 2])))', 1),
    ('(let (cdr (fun (pair) (select [(case [(,car) (,cdr)] pair) cdr] (" error)))) (cdr (" [1 2])))', 2),
]

if __name__ == "__main__":
    for (tutorial, answer) in tutorials:
        print(tutorial)
        [o] = run(read(tutorial))
        if o != answer:
            print(f'expected {answer}')
        print(f'found {o}')
        print()
